//############################################################################//
//Made in 2010-2015 by Artyom Litvinovich
//Based on Petit FAT FS
//############################################################################//
#include <avr/io.h>
//############################################################################//
#define CS_LOW()  clear_bit(SD_PORT,SD_CS) //CS=L
#define CS_HIGH()   set_bit(SD_PORT,SD_CS) //CS=H

#define STA_NOINIT  0x01 //Drive not initialized
#define STA_NODISK  0x02 //No medium in the drive

//Definitions for MMC/SDC command 
#define CMD0   (0x40+0)  //GO_IDLE_STATE
#define CMD1   (0x40+1)  //SEND_OP_COND(MMC)
#define ACMD41 (0xC0+41) //SEND_OP_COND(SDC)
#define ACMD23 (0xC0+23) //SET_WR_BLK_ERASE_COUNT (SDC)
#define CMD8   (0x40+8)  //SEND_IF_COND
#define CMD12  (0x40+12) //STOP_TRANSMISSION
#define CMD16  (0x40+16) //SET_BLOCKLEN
#define CMD17  (0x40+17) //READ_SINGLE_BLOCK
#define CMD18  (0x40+18) //READ_MULTIPLE_BLOCK
#define CMD24  (0x40+24) //WRITE_BLOCK
#define CMD25  (0x40+25)  //WRITE_MULTIPLE_BLOCK
#define CMD55  (0x40+55) //APP_CMD
#define CMD58  (0x40+58) //READ_OCR
#define CMD9   (9)       //Send CSD
//############################################################################//
//Card type flags
#define CT_MMC   0x01 //MMC ver 3
#define CT_SD1   0x02 //SD ver 1
#define CT_SD2   0x04 //SD ver 2
#define CT_BLOCK 0x08 //Block addressing
//############################################################################//
#define RES_OK 0  //0: Function succeeded
#define RES_ERROR 1 //1: Disk error
#define RES_NOTRDY 2  //2: Not ready
#define RES_PARERR 3  //3: Invalid parameter
//############################################################################//
#define GET_SECTOR_COUNT 1
#define CTRL_SYNC 2
//############################################################################//
#define mmc_spi_on SPCR=_BV(SPE)|_BV(MSTR)|_BV(SPR0);SPSR=0x01
//############################################################################//
static byte card_type;
static byte mmc_inited=0;
//############################################################################//
byte xchg_spi(byte dat)
{
 SPDR=dat;
 loop_until_bit_is_set(SPSR,SPIF);
 return SPDR;
}
//############################################################################//
void xmit_spi_multi(const byte *p,word cnt)
{
 do{
  SPDR=*p++;loop_until_bit_is_set(SPSR,SPIF);
  SPDR=*p++;loop_until_bit_is_set(SPSR,SPIF);
 }while(cnt-=2);
}
//############################################################################//
void rcvr_spi_multi(byte *p,word cnt)
{
 do{
  SPDR=0xFF;loop_until_bit_is_set(SPSR,SPIF);*p++ =SPDR;
  SPDR=0xFF;loop_until_bit_is_set(SPSR,SPIF);*p++ =SPDR;
 }while(cnt-=2);
}
//############################################################################//
int wait_ready(void) 
{
 byte d;
 word t;

 t=25000;
 do{
  d=xchg_spi(0xFF);
 }while(d!=0xFF&&(t--));

 return(d==0xFF)?1:0;
}
//############################################################################//
int rcvr_datablock(byte  *buff,word btr)
{
 byte token;
 word t;

 //Wait for data packet in timeout of 200ms
 t=200;
 do{
  token=xchg_spi(0xFF);
  if(token==0xFF)_delay_ms(1);
 }while((token==0xFF)&&(t--));
 
 if(token!=0xFE)return 0; //If not valid data token, retutn with error

 rcvr_spi_multi(buff, btr);  //Receive the data block into buffer
 xchg_spi(0xFF);     //Discard CRC
 xchg_spi(0xFF);

 return 1;      //Return with success
}
//############################################################################//
int xmit_datablock(const byte *buff,byte token)
{
 byte resp;

 //if (!wait_ready(500)) return 0;
 if (!wait_ready()) return 0;

 xchg_spi(token);     //Xmit data token
 if (token!=0xFD){ //Is data token
  xmit_spi_multi(buff,512);  //Xmit the data block to the MMC
  xchg_spi(0xFF);     //CRC (Dummy)
  xchg_spi(0xFF);
  resp=xchg_spi(0xFF);   //Reveive data response
  if ((resp&0x1F)!=0x05)return 0;  //If not accepted, return with error
 }
 return 1;
}
//############################################################################//
void deselect(void)
{
 mmc_spi_on;
 
 CS_HIGH();
 xchg_spi(0xFF);
}
//############################################################################//
int select(void)
{
 mmc_spi_on;
 
 CS_LOW();
 xchg_spi(0xFF);

 if (wait_ready())return 1;
 deselect();
 return 1;
}
//############################################################################//
//Send a command packet to MMC
//1st byte(Start + Index)
//Argument(32 bits)
byte send_cmd(byte cmd,dword arg)
{
 byte n,res;

 //ACMD<n> is the command sequense of CMD55-CMD<n>
 if (cmd&0x80){
  cmd&=0x7F;
  res=send_cmd(CMD55,0);
  if(res>1)return res;
 }

 //Select the card
 if (cmd!=CMD12){
  deselect();
  if (!select())return 0xFF;
 }

 //Send a command packet
 xchg_spi (0x40 | cmd);      //Start + Command index
 xchg_spi ((byte)(arg >> 24));  //Argument[31..24]
 xchg_spi ((byte)(arg >> 16));  //Argument[23..16]
 xchg_spi ((byte)(arg >> 8));   //Argument[15..8]
 xchg_spi ((byte)arg);    //Argument[7..0]
 n=0x01;       //Dummy CRC + Stop
 if(cmd==CMD0)n=0x95;   //Valid CRC for CMD0(0)
 if(cmd==CMD8)n=0x87;   //Valid CRC for CMD8(0x1AA)
 xchg_spi(n);

 //Receive a command response
 if(cmd==CMD12)xchg_spi(0xFF);
 n=10;        //Wait for a valid response in timeout of 10 attempts
 do{
  res=xchg_spi(0xFF);
 }while((res & 0x80)&& --n);

 return res;   //Return with the response value
}
//############################################################################//
//Initialize Disk Drive
byte disk_initialize(byte redo)
{
 byte n,cmd,ty,ocr[4];
 word tmr;

 if (mmc_inited && !redo)return 0;
 mmc_inited=0;

 deselect();
 for(n=10;n;n--)xchg_spi(0xFF); //80 dummy clocks with CS=H

 ty=0;
 if (send_cmd(CMD0,0)==1){   //Enter Idle state
  if(send_cmd(CMD8,0x1AA)==1){ //SDv2
   for(n=0;n<4;n++)ocr[n]=xchg_spi(0xFF);  //Get trailing return value of R7 resp 
   if(ocr[2]==0x01 && ocr[3]==0xAA){   //The card can work at vdd range of 2.7-3.6V 
    for(tmr=10000;tmr && send_cmd(ACMD41,1UL<<30);tmr--)_delay_us(100); //Wait for leaving idle state(ACMD41 with HCS bit)
    if(tmr && send_cmd(CMD58,0)== 0){  //Check CCS bit in the OCR 
     for(n=0; n<4; n++)ocr[n]=xchg_spi(0xFF);
     ty=(ocr[0] & 0x40)? CT_SD2 | CT_BLOCK : CT_SD2; //SDv2(HC or SC)
    }
   }
  }else{       //SDv1 or MMCv3 
   if(send_cmd(ACMD41,0)<= 1) {
    ty=CT_SD1; cmd=ACMD41; //SDv1 
   }else{
    ty=CT_MMC; cmd=CMD1; //MMCv3
   }
   for(tmr=10000; tmr && send_cmd(cmd,0); tmr--)_delay_us(100); //Wait for leaving idle state
   if(!tmr || send_cmd(CMD16,512)!= 0)   //Set R/W block length to 512
    ty=0;
  }
 }
 card_type=ty;
 deselect();

 if (ty)mmc_inited=1;
 return ty?0:STA_NOINIT;
}
//############################################################################//
byte disk_ioctl(byte ctrl,void *buff)
{
 byte res;
 byte n,csd[16];
 dword csize;

 res=RES_ERROR;


 switch (ctrl){
  //Get number of sectors on the disk (dword) 
  case GET_SECTOR_COUNT:{
   if((send_cmd(CMD9,0)==0) && rcvr_datablock(csd, 16)) {
    if ((csd[0] >> 6) == 1) { //SDC ver 2.00 
     csize=csd[9] + ((word)csd[8] << 8) + ((dword)(csd[7] & 63) << 16) + 1;
     *(dword*)buff=csize << 10;
    } else {     //SDC ver 1.XX or MMC
     n=(csd[5] & 15) + ((csd[10] & 128) >> 7) + ((csd[9] & 3) << 1) + 2;
     csize=(csd[8] >> 6) + ((word)csd[7] << 2) + ((word)(csd[6] & 3) << 10) + 1;
     *(dword*)buff=csize << (n - 9);
    }
    res=RES_OK;
   }
   break;
  }
  //Make sure that no pending write process. Do not remove this or written sector might not left updated.
  case CTRL_SYNC:
   if (select()) res=RES_OK;
   break;
/*
 case GET_BLOCK_SIZE : //Get erase block size in unit of sector (dword) 
  if (CardType & CT_SD2) { //SDv2? 
   if (send_cmd(ACMD13, 0) == 0) { //Read SD status 
    xchg_spi(0xFF);
    if (rcvr_datablock(csd, 16)) {    //Read partial block 
     for (n=64 - 16; n; n--) xchg_spi(0xFF); //Purge trailing data 
     *(dword*)buff=16UL << (csd[10] >> 4);
     res=RES_OK;
    }
   }
  } else {     //SDv1 or MMCv3 
   if ((send_cmd(CMD9, 0) == 0) && rcvr_datablock(csd, 16)) { //Read CSD 
    if (CardType & CT_SD1) { //SDv1 
     *(dword*)buff=(((csd[10] & 63) << 1) + ((word)(csd[11] & 128) >> 7) + 1) << ((csd[13] >> 6) - 1);
    } else {     //MMCv3 
     *(dword*)buff=((word)((csd[10] & 124) >> 2) + 1) * (((csd[11] & 3) << 3) + ((csd[11] & 224) >> 5) + 1);
    }
    res=RES_OK;
   }
  }
  break;


 case MMC_GET_CSD :  //Receive CSD as a data block (16 bytes) 
  if (send_cmd(CMD9, 0) == 0  //READ_CSD 
   && rcvr_datablock(ptr, 16))
   res=RES_OK;
  break;

 case MMC_GET_CID :  //Receive CID as a data block (16 bytes) 
  if (send_cmd(CMD10, 0) == 0  //READ_CID 
   && rcvr_datablock(ptr, 16))
   res=RES_OK;
  break;

 case MMC_GET_OCR :  //Receive OCR as an R3 resp (4 bytes) 
  if (send_cmd(CMD58, 0) == 0) { //READ_OCR 
   for (n=4; n; n--) *ptr++=xchg_spi(0xFF);
   res=RES_OK;
  }
  break;

 case MMC_GET_SDSTAT : //Receive SD statsu as a data block (64 bytes) 
  if (send_cmd(ACMD13, 0) == 0) { //SD_STATUS 
   xchg_spi(0xFF);
   if (rcvr_datablock(ptr, 64))
    res=RES_OK;
  }
  break;
*/
  default:res=RES_PARERR;
 }
 deselect();
 return res;
}
//############################################################################//
//Read partial sector
byte disk_readp(byte *buff,dword lba,word ofs,word cnt)
{
 byte res;
 byte rc;
 word bc;

 if (!(card_type&CT_BLOCK))lba*=512;

 res=RES_ERROR;
 if (!buff)return res;
  
 if (send_cmd(CMD17,lba)==0){ 
  bc=40000;
  do{       //Wait for data packet
   rc=xchg_spi(0xFF);
  }while(rc == 0xFF && --bc);

  if(rc==0xFE){    //A data packet arrived
   bc=514-ofs-cnt;

   //Skip leading bytes
   if(ofs){
    do xchg_spi(0xFF); while(--ofs);
   }

   //Receive a part of the sector
   do{
    *buff++=xchg_spi(0xFF);
   }while(--cnt);
  

   //Skip trailing bytes and CRC 
   do xchg_spi(0xFF); while(--bc);

   res=RES_OK;
  }
 }

 deselect();

 return res;
}
//############################################################################//
byte disk_read(byte *buff,dword sector,word count)
{
 byte res;
 res=RES_ERROR;
 if (!buff)return res;
 if (!(card_type&CT_BLOCK))sector*=512;

 if (count==1){ //Single block read
  if ((send_cmd(CMD17,sector)==0)&& rcvr_datablock(buff,512))count=0;
 }else{         //Multiple block read
  if (send_cmd(CMD18,sector)==0){ //READ_MULTIPLE_BLOCK
   do{
    if (!rcvr_datablock(buff,512)) break;
    buff+=512;
   }while(--count);
   send_cmd(CMD12,0);    //STOP_TRANSMISSION
  }
 }
 deselect();

 return count ? RES_ERROR : RES_OK;
}
//############################################################################//
byte disk_writep(const byte *buff, dword sa)
{
 byte res;
 word bc;
 static word wc;

 res=RES_ERROR;

 if (buff){  //Send data bytes
  bc=(word)sa;
  while (bc && wc){  //Send data bytes to the card
   xchg_spi(*buff++);
   wc--; 
   bc--;
  }
  res=RES_OK;
 }else{
  if (sa){ //Initiate sector write process
   if (!(card_type & CT_BLOCK))sa*=512; //Convert to byte address if needed
   if (send_cmd(CMD24, sa)==0){   //WRITE_SINGLE_BLOCK
    xchg_spi(0xFF); 
    xchg_spi(0xFE);  //Data block header
    wc=512;       //Set byte counter
    res=RES_OK;
   }
  }else{ //Finalize sector write process
   bc=wc + 2;
   while (bc--) xchg_spi(0); //Fill left bytes and CRC with zeros 
   if ((xchg_spi(0xFF) & 0x1F)==0x05) { //Receive data resp and wait for end of write process in timeout of 500ms
    for (bc=5000; xchg_spi(0xFF)!=0xFF && bc;bc--) _delay_us(100); //Wait ready
    if (bc)res=RES_OK;
   }
   deselect();
   xchg_spi(0xFF);
  }
 }

 return res;
}
//############################################################################//
byte disk_write(const byte *buff,dword sector,word count)
{
 //byte res;
 if (!(card_type & CT_BLOCK))sector*=512; //Convert to byte address if needed

 if (count==1){ //Single block write
  if ((send_cmd(CMD24,sector)==0)&& xmit_datablock(buff,0xFE))count=0;
 }else{    //Multiple block write
  if (card_type&(CT_SD1|CT_SD2))send_cmd(ACMD23,count);
  if (send_cmd(CMD25, sector)==0){ //WRITE_MULTIPLE_BLOCK
   do{
    if (!xmit_datablock(buff,0xFC))break;
    buff+=512;
   }while(--count);
   if(!xmit_datablock(0,0xFD))count=1;
  }
 }
 deselect();

 return count ? RES_ERROR : RES_OK;
}
//############################################################################//
