//############################################################################//
#include <avr/io.h>
//############################################################################//
#define SD_CS 0

#define CS_LOW()  clear_bit(PORTC,SD_CS) //CS=L
#define CS_HIGH()   set_bit(PORTC,SD_CS) //CS=H
#define mmc_spi_on SPCR=_BV(SPE)|_BV(MSTR)|_BV(SPR0);SPSR=0x01

#define STA_NOINIT  0x01 //Drive not initialized
#define STA_NODISK  0x02 //No medium in the drive

//Definitions for MMC/SDC command 
#define CMD0   (0x40+0)  //GO_IDLE_STATE
#define CMD1   (0x40+1)  //SEND_OP_COND(MMC)
#define ACMD41 (0xC0+41) //SEND_OP_COND(SDC)
#define CMD8   (0x40+8)  //SEND_IF_COND
#define CMD12  (12)  //STOP_TRANSMISSION
#define CMD16  (0x40+16) //SET_BLOCKLEN
#define CMD17  (0x40+17) //READ_SINGLE_BLOCK
#define CMD24  (0x40+24) //WRITE_BLOCK
#define CMD55  (0x40+55) //APP_CMD
#define CMD58  (0x40+58) //READ_OCR
#define CMD9   (9)       //Send CSD
//############################################################################//
//Card type flags
#define CT_MMC   0x01 //MMC ver 3
#define CT_SD1   0x02 //SD ver 1
#define CT_SD2   0x04 //SD ver 2
#define CT_BLOCK 0x08 //Block addressing
//############################################################################//
#define RES_OK 0  //0: Function succeeded
#define RES_ERROR 1 //1: Disk error
#define RES_NOTRDY 2  //2: Not ready
#define RES_PARERR 3  //3: Invalid parameter
//############################################################################//
#define GET_SECTOR_COUNT 1
//############################################################################//
static byte card_type;
static byte mmc_inited=0;
//############################################################################//
byte xchg_spi(byte dat)
{
 SPDR=dat;
 loop_until_bit_is_set(SPSR,SPIF);
 return SPDR;
}
//############################################################################//
void rcvr_spi_multi(byte *p,word cnt)
{
 do{
  SPDR=0xFF;loop_until_bit_is_set(SPSR,SPIF);*p++ =SPDR;
  SPDR=0xFF;loop_until_bit_is_set(SPSR,SPIF);*p++ =SPDR;
 }while(cnt-=2);
}
//############################################################################//
int wait_ready(void) 
{
 byte d;
 word t;

 t=25000;
 do{
  d=xchg_spi(0xFF);
 }while(d!=0xFF&&(t--));

 return(d==0xFF)?1:0;
}
//############################################################################//
void deselect(void)
{
 mmc_spi_on;
 CS_HIGH();
 xchg_spi(0xFF);
}
//############################################################################//
int select(void)
{
 mmc_spi_on;
 CS_LOW();
 xchg_spi(0xFF);

 if (wait_ready())return 1;
 deselect();
 return 1;
}
//############################################################################//
//Send a command packet to MMC
//1st byte(Start + Index)
//Argument(32 bits)
byte send_cmd(byte cmd,dword arg)
{
 byte n,res;

 //ACMD<n> is the command sequense of CMD55-CMD<n>
 if (cmd&0x80){
  cmd&=0x7F;
  res=send_cmd(CMD55,0);
  if(res>1)return res;
 }

 //Select the card
 deselect();
 if (!select())return 0xFF;

 //Send a command packet
 xchg_spi (0x40 | cmd);      //Start + Command index
 xchg_spi ((byte)(arg >> 24));  //Argument[31..24]
 xchg_spi ((byte)(arg >> 16));  //Argument[23..16]
 xchg_spi ((byte)(arg >> 8));   //Argument[15..8]
 xchg_spi ((byte)arg);    //Argument[7..0]
 n=0x01;       //Dummy CRC + Stop
 if(cmd==CMD0)n=0x95;   //Valid CRC for CMD0(0)
 if(cmd==CMD8)n=0x87;   //Valid CRC for CMD8(0x1AA)
 xchg_spi(n);

 //Receive a command response
 if(cmd==CMD12)xchg_spi(0xFF);
 n=10;        //Wait for a valid response in timeout of 10 attempts
 do{
  res=xchg_spi(0xFF);
 }while((res & 0x80)&& --n);

 return res;   //Return with the response value
}
//############################################################################//
//Initialize Disk Drive
byte disk_initialize(void)
{
 byte n,cmd,ty,ocr[4];
 word tmr;

 if (mmc_inited)return 0;

 deselect();
 for(n=10;n;n--)xchg_spi(0xFF); //80 dummy clocks with CS=H

 ty=0;
 if (send_cmd(CMD0,0)==1){   //Enter Idle state
  if(send_cmd(CMD8,0x1AA)==1){ //SDv2
   for(n=0;n<4;n++)ocr[n]=xchg_spi(0xFF);  //Get trailing return value of R7 resp 
   if(ocr[2]==0x01 && ocr[3]==0xAA){   //The card can work at vdd range of 2.7-3.6V 
    for(tmr=10000;tmr && send_cmd(ACMD41,1UL<<30);tmr--)_delay_us(100); //Wait for leaving idle state(ACMD41 with HCS bit)
    if(tmr && send_cmd(CMD58,0)== 0){  //Check CCS bit in the OCR 
     for(n=0; n<4; n++)ocr[n]=xchg_spi(0xFF);
     ty=(ocr[0] & 0x40)? CT_SD2 | CT_BLOCK : CT_SD2; //SDv2(HC or SC)
    }
   }
  }else{       //SDv1 or MMCv3 
   if(send_cmd(ACMD41,0)<= 1) {
    ty=CT_SD1; cmd=ACMD41; //SDv1 
   }else{
    ty=CT_MMC; cmd=CMD1; //MMCv3
   }
   for(tmr=10000; tmr && send_cmd(cmd,0); tmr--)_delay_us(100); //Wait for leaving idle state
   if(!tmr || send_cmd(CMD16,512)!= 0)   //Set R/W block length to 512
    ty=0;
  }
 }
 card_type=ty;
 deselect();

 if (ty)mmc_inited=1;
 return ty?0:STA_NOINIT;
}
//############################################################################//
//Read partial sector
byte disk_readp(byte *buff,dword lba,word ofs,word cnt)
{
 byte res;
 byte rc;
 word bc;

 if (!(card_type&CT_BLOCK))lba*=512;

 res=RES_ERROR;
 if (!buff)return res;
  
 if (send_cmd(CMD17,lba)==0){ 
  bc=40000;
  do{       //Wait for data packet
   rc=xchg_spi(0xFF);
  }while(rc == 0xFF && --bc);

  if(rc==0xFE){    //A data packet arrived
   bc=514-ofs-cnt;

   //Skip leading bytes
   if(ofs){
    do xchg_spi(0xFF); while(--ofs);
   }

   //Receive a part of the sector
   do{
    *buff++=xchg_spi(0xFF);
   }while(--cnt);
  

   //Skip trailing bytes and CRC 
   do xchg_spi(0xFF); while(--bc);

   res=RES_OK;
  }
 }

 deselect();

 return res;
}
//############################################################################//
