//############################################################################//
#define _WORD_ACCESS 1
/* The _WORD_ACCESS option defines which access method is used to the word
/  data in the FAT structure.
/
/   0: Byte-by-byte access. Always compatible with all platforms.
/   1: Word access. Do not choose this unless following condition is met.
/
/  When the byte order on the memory is big-endian or address miss-aligned
/  word access results incorrect behavior, the _WORD_ACCESS must be set to 0.
/  If it is not the case, the value can also be set to 1 to improve the
/  performance and code efficiency. */
//############################################################################//
#define CLUST word
//############################################################################//
//File system object structure
typedef struct{
 byte flag;        //File status flags 
 byte csize;       //Number of sectors per cluster 
 byte pad1;
 word n_rootdir;   //Number of root directory entries (0 on FAT32) 
 CLUST n_fatent;   //Number of FAT entries (= number of clusters + 2) 
 dword fatbase;    //FAT start sector 
 dword dirbase;    //Root directory start sector (Cluster# on FAT32) 
 dword database;   //Data start sector 
 dword fptr;       //File R/W pointer 
 dword fsize;      //File size 
 CLUST org_clust;  //File start cluster 
 CLUST curr_clust; //File current cluster 
 dword dsect;      //File current data sector 
}FATFS;
//############################################################################//
//Directory object structure
typedef struct{
 word index;    //Current read/write index number 
 byte* fn;      //Pointer to the SFN (in/out) {file[8],ext[3],status[1]} 
 CLUST sclust;  //Table start cluster (0:Static table) 
 CLUST clust;   //Current cluster 
 dword sect;    //Current sector 
}DIR;
//############################################################################//
//File status structure
typedef struct {
 dword fsize;    //File size 
 word fdate;     //Last modified date 
 word ftime;     //Last modified time 
 byte fattrib;   //Attribute 
 char fname[13]; //File name 
}FILINFO;
//############################################################################//
//File function return code
#define FR_OK            0
#define FR_DISK_ERR      1
#define FR_NOT_READY     2
#define FR_NO_FILE       3
#define FR_NO_PATH       4
#define FR_NOT_OPENED    5
#define FR_NOT_ENABLED   6
#define FR_NO_FILESYSTEM 7
//############################################################################//
//File status flag (FATFS.flag)
#define FA_OPENED 0x01
#define FA_WPRT   0x02
#define FA__WIP   0x40
//############################################################################//
//File attribute bits for directory entry
#define AM_RDO  0x01 //Read only 
#define AM_HID  0x02 //Hidden 
#define AM_SYS  0x04 //System 
#define AM_VOL  0x08 //Volume label 
#define AM_LFN  0x0F //LFN entry 
#define AM_DIR  0x10 //Directory 
#define AM_ARC  0x20 //Archive 
#define AM_MASK 0x3F //Mask of defined bits 
//############################################################################//
//Multi-byte word access macros
#if _WORD_ACCESS == 1 //Enable word access to the FAT structure 
#define LD_word(ptr)  (word)(*(word*)(byte*)(ptr))
#define LD_dword(ptr)  (dword)(*(dword*)(byte*)(ptr))
#define ST_word(ptr,val) *(word*)(byte*)(ptr)=(word)(val)
#define ST_dword(ptr,val) *(dword*)(byte*)(ptr)=(dword)(val)
#else     //Use byte-by-byte access to the FAT structure 
#define LD_word(ptr)  (word)(((word)*((byte*)(ptr)+1)<<8)|(word)*(byte*)(ptr))
#define LD_dword(ptr)  (dword)(((dword)*((byte*)(ptr)+3)<<24)|((dword)*((byte*)(ptr)+2)<<16)|((word)*((byte*)(ptr)+1)<<8)|*(byte*)(ptr))
#define ST_word(ptr,val) *(byte*)(ptr)=(byte)(val); *((byte*)(ptr)+1)=(byte)((word)(val)>>8)
#define ST_dword(ptr,val) *(byte*)(ptr)=(byte)(val); *((byte*)(ptr)+1)=(byte)((word)(val)>>8); *((byte*)(ptr)+2)=(byte)((dword)(val)>>16); *((byte*)(ptr)+3)=(byte)((dword)(val)>>24)
#endif
//############################################################################//
#define LD_CLUST(dir)LD_word(dir+DIR_FstClusLO)

#define _EXCVT { \
 0x80,0x81,0x82,0x82,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x80,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x8A,0x9B,0x8C,0x8D,0x8E,0x8F,\
 0xA0,0xA2,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,0xB0,0xB1,0xB2,0xB2,0xA5,0xB5,0xB6,0xB7,0xA8,0xB9,0xAA,0xBB,0xA3,0xBD,0xBD,0xAF,\
 0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF,\
 0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF  \
}
//############################################################################//
//Character code support macros
#define IsUpper(c)(((c)>='A')&&((c)<='Z'))
#define IsLower(c)(((c)>='a')&&((c)<='z'))

#define IsDBCS1(c)0
#define IsDBCS2(c)0
//############################################################################//
//FatFs refers the members in the FAT structures with byte offset instead
//of structure member because there are incompatibility of the packing option
//between various compilers.

#define BS_jmpBoot   0
#define BS_OEMName   3
#define BPB_BytsPerSec  11
#define BPB_SecPerClus  13
#define BPB_RsvdSecCnt  14
#define BPB_NumFATs   16
#define BPB_RootEntCnt  17
#define BPB_TotSec16  19
#define BPB_Media   21
#define BPB_FATSz16   22
#define BPB_SecPerTrk  24
#define BPB_NumHeads  26
#define BPB_HiddSec   28
#define BPB_TotSec32  32
#define BS_55AA    510

#define BS_DrvNum   36
#define BS_BootSig   38
#define BS_VolID   39
#define BS_VolLab   43
#define BS_FilSysType  54

#define BPB_FATSz32   36
#define BPB_ExtFlags  40
#define BPB_FSVer   42
#define BPB_RootClus  44
#define BPB_FSInfo   48
#define BPB_BkBootSec  50
#define BS_DrvNum32   64
#define BS_BootSig32  66
#define BS_VolID32   67
#define BS_VolLab32   71
#define BS_FilSysType32  82

#define MBR_Table   446

#define DIR_Name   0
#define DIR_Attr   11
#define DIR_NTres   12
#define DIR_CrtTime   14
#define DIR_CrtDate   16
#define DIR_FstClusHI  20
#define DIR_WrtTime   22
#define DIR_WrtDate   24
#define DIR_FstClusLO  26
#define DIR_FileSize  28
//############################################################################//
//Pointer to the file system object(logical drive)
static FATFS *FatFs;
//############################################################################//
//Fill memory
static void mem_set(void* dst,int val,int cnt)
{
 char *d=(char*)dst;
 while(cnt--)*d++=(char)val;
}
//############################################################################//
//Compare memory to memory
static int mem_cmp(const void* dst,const void* src,int cnt)
{
 const char *d=(const char *)dst,*s=(const char *)src;
 int r=0;
 while(cnt-- &&(r=*d++-*s++)==0);
 return r;
}
//############################################################################//
//FAT access-Read value of a FAT entry
static CLUST get_fat(CLUST clst)
{
 byte buf[4];
 FATFS *fs=FatFs;

 if(clst<2 || clst>=fs->n_fatent)return 1;
  
 if(disk_readp(buf,fs->fatbase+clst/256,(word)(((word)clst%256)*2),2))return 1;
 return LD_word(buf);
}
//############################################################################//
//Get sector# from cluster #
//!=0: Sector number,0: Failed-invalid cluster #
static dword clust2sect(CLUST clst)
{
 FATFS *fs=FatFs;
 clst -= 2;
 if(clst >=(fs->n_fatent-2))return 0;  //Invalid cluster #
 return (dword)clst*fs->csize+fs->database;
}
//############################################################################//
//Directory handling-Rewind directory index
static byte dir_rewind(DIR *dj)
{
 CLUST clst;
 FATFS *fs=FatFs;


 dj->index=0;
 clst=dj->sclust;
 if(clst==1 || clst >= fs->n_fatent)return FR_DISK_ERR; //Check start cluster range
 dj->clust=clst;      //Current cluster
 dj->sect=clst ? clust2sect(clst): fs->dirbase; //Current sector

 return FR_OK; //Seek succeeded
}
//############################################################################//
//Directory handling-Move directory index next
//FR_OK: Succeeded
//FR_NO_FILE: End of table
static byte dir_next(DIR *dj)
{
 CLUST clst;
 word i;
 FATFS *fs=FatFs;


 i=dj->index+1;
 if(!i || !dj->sect)/* Report EOT when index has reached 65535 */
  return FR_NO_FILE;

 if(!(i % 16)){  /* Sector changed? */
  dj->sect++;   /* Next sector */

  if(dj->clust==0){ /* Static table */
   if(i >= fs->n_rootdir)/* Report EOT when end of table */
    return FR_NO_FILE;
  }
  else{     /* Dynamic table */
   if(((i / 16)&(fs->csize-1))==0){ /* Cluster changed? */
    clst=get_fat(dj->clust);  /* Get next cluster */
    if(clst <= 1)return FR_DISK_ERR;
    if(clst >= fs->n_fatent) /* When it reached end of dynamic table */
     return FR_NO_FILE;   /* Report EOT */
    dj->clust=clst;    /* Initialize data for new cluster */
    dj->sect=clust2sect(clst);
   }
  }
 }

 dj->index=i;

 return FR_OK;
}




/*-----------------------------------------------------------------------*/
/* Directory handling-Find an object in the directory                  */
/*-----------------------------------------------------------------------*/

static
byte dir_find(
 DIR *dj, /* Pointer to the directory object linked to the file name */
 byte *dir  /* 32-byte working buffer */
)
{
 byte res;
 byte c;


 res=dir_rewind(dj);   /* Rewind directory object */
 if(res != FR_OK)return res;

 do{
  res=disk_readp(dir,dj->sect,(word)((dj->index % 16)* 32),32)/* Read an entry */
   ? FR_DISK_ERR : FR_OK;
  if(res != FR_OK)break;
  c=dir[DIR_Name]; /* First character */
  if(c==0){ res=FR_NO_FILE; break; }/* Reached to end of table */
  if(!(dir[DIR_Attr] & AM_VOL)&& !mem_cmp(dir,dj->fn,11))/* Is it a valid entry? */
   break;
  res=dir_next(dj);     /* Next entry */
 }while(res==FR_OK);

 return res;
}
//############################################################################//


/* Pick a segment and create the object name in directory form           */
/*-----------------------------------------------------------------------*/

#ifdef _EXCVT
 static const byte cvt[]=_EXCVT;
#endif

static
byte create_name(
 DIR *dj,  /* Pointer to the directory object */
 const char **path /* Pointer to pointer to the segment in the path string */
)
{
 byte c,d,ni,si,i,*sfn;
 const char *p;

 /* Create file name in directory form */
 sfn=dj->fn;
 mem_set(sfn,' ',11);
 si=i=0; ni=8;
 p=*path;
 for(;;){
  c=p[si++];
  if(c <= ' ' || c=='/')break; /* Break on end of segment */
  if(c=='.' || i >= ni){
   if(ni != 8 || c != '.')break;
   i=8; ni=11;
   continue;
  }
#ifdef _EXCVT
  if(c >= 0x80)    /* To upper extended char(SBCS)*/
   c=cvt[c-0x80];
#endif
  if(IsDBCS1(c)&& i<ni-1){ /* DBC 1st byte? */
   d=p[si++];    /* Get 2nd byte */
   sfn[i++]=c;
   sfn[i++]=d;
  }else{      /* Single byte code */
   if(IsLower(c))c -= 0x20; /* toupper */
   sfn[i++]=c;
  }
 }
 *path=&p[si];      /* Rerurn pointer to the next segment */

 sfn[11]=(c <= ' ')? 1 : 0;  /* Set last segment flag if end of path */

 return FR_OK;
}


/*-----------------------------------------------------------------------*/
/* Follow a file path                                                    */
/*-----------------------------------------------------------------------*/

static
byte follow_path( /* FR_OK(0): successful,!=0: error code */
 DIR *dj,  /* Directory object to return last directory and found object */
 byte *dir,  /* 32-byte working buffer */
 const char *path /* Full-path string to find a file or directory */
)
{
 byte res;


 while(*path==' ')path++;  /* Skip leading spaces */
 if(*path=='/')path++;   /* Strip heading separator */
 dj->sclust=0;      /* Set start directory(always root dir)*/

 if ((byte)*path <= ' '){   /* Null path means the root directory */
  res=dir_rewind(dj);
  dir[0]=0;

 }else{       /* Follow path */
  for(;;){
   res=create_name(dj,&path); /* Get a segment */
   if(res != FR_OK)break;
   res=dir_find(dj,dir);  /* Find it */
   if(res != FR_OK){    /* Could not find the object */
    if(res==FR_NO_FILE && !*(dj->fn+11))
     res=FR_NO_PATH;
    break;
   }
   if(*(dj->fn+11))break;  /* Last segment match. Function completed. */
   if(!(dir[DIR_Attr] & AM_DIR)){ /* Cannot follow because it is a file */
    res=FR_NO_PATH; break;
   }
   dj->sclust=LD_CLUST(dir);
  }
 }

 return res;
}
//############################################################################//
//Check a sector if it is an FAT boot record
//0: The FAT boot record
//1: Valid boot record but not an FAT
//2: Not a boot record
//3: Error
static byte check_fs(byte *buf,dword sect)
{
 //Read the boot sector
 if (disk_readp(buf,sect,510,2))return 3;
 //Check record signature
 if (LD_word(buf)!=0xAA55)return 2;
 //Check FAT116
 if (!disk_readp(buf,sect,BS_FilSysType,2)&& LD_word(buf)==0x4146)return 0;
 return 1;
}
//############################################################################//
//############################################################################//
//Mount/Unmount a Locical Drive
byte pf_mount(FATFS *fs)
{
 byte fmt,buf[36];
 dword bsect,fsize,tsect,mclst;

 FatFs=0;
 if (!fs)return FR_OK;    //Unregister fs object

 if (disk_initialize()& STA_NOINIT)return FR_NOT_READY; //Check if the drive is ready or not

 //Search FAT partition on the drive
 bsect=0;
 fmt=check_fs(buf,bsect);   //Check sector 0 as an SFD format
 if(fmt==1){      //Not an FAT boot record,it may be FDISK format
  //Check a partition listed in top of the partition table
  if(disk_readp(buf,bsect,MBR_Table,16)){ //1st partition entry
   fmt=3;
  }else if(buf[4]){     //Is the partition existing?
   bsect=LD_dword(&buf[8]); //Partition offset in LBA
   fmt=check_fs(buf,bsect); //Check the partition
  }
 }
 if(fmt==3)return FR_DISK_ERR;
 if (fmt)return FR_NO_FILESYSTEM;  //No valid FAT patition is found

 //Initialize the file system object
 if (disk_readp(buf,bsect,13,sizeof(buf)))return FR_DISK_ERR;

 fsize=LD_word(buf+BPB_FATSz16-13);    //Number of sectors per FAT
 if (!fsize)fsize=LD_dword(buf+BPB_FATSz32-13);

 fsize*=buf[BPB_NumFATs-13];      //Number of sectors in FAT area
 fs->fatbase=bsect+LD_word(buf+BPB_RsvdSecCnt-13); //FAT start sector(lba)
 fs->csize=buf[BPB_SecPerClus-13];     //Number of sectors per cluster 
 fs->n_rootdir=LD_word(buf+BPB_RootEntCnt-13);  //Nmuber of root directory entries
 tsect=LD_word(buf+BPB_TotSec16-13);    //Number of sectors on the file system
 if (!tsect)tsect=LD_dword(buf+BPB_TotSec32-13);
 mclst=(tsect-LD_word(buf+BPB_RsvdSecCnt-13)-fsize-fs->n_rootdir/16)/fs->csize+2;
 fs->n_fatent=(CLUST)mclst;

 if(mclst<0xFF7)return FR_NO_FILESYSTEM;      //Number of clusters<0xFF5
 if(mclst>=0xFFF7)return FR_NO_FILESYSTEM;    //Number of clusters>=0xFFF5

 fs->dirbase=fs->fatbase+fsize;    //Root directory start sector(lba)
 fs->database=fs->fatbase+fsize+fs->n_rootdir/16; //Data start sector(lba)

 fs->flag=0;
 FatFs=fs;

 return FR_OK;
}
//############################################################################//
//Open or Create a File
byte pf_open(const char *path)
{
 byte res;
 DIR dj;
 byte sp[12],dir[32];
 FATFS *fs=FatFs;

 if (!fs)return FR_NOT_ENABLED;     //Check file system

 fs->flag=0;
 dj.fn=sp;
 res=follow_path(&dj,dir,path);                           //Follow the file path
 if(res!=FR_OK)return res;                                //Follow failed
 if(!dir[0] ||(dir[DIR_Attr] & AM_DIR))return FR_NO_FILE; //It is a directory

 fs->org_clust=LD_CLUST(dir);          //File start cluster
 fs->fsize=LD_dword(dir+DIR_FileSize); //File size
 fs->fptr=0;                           //File pointer
 fs->flag=FA_OPENED;

 return FR_OK;
}
//############################################################################//
//Read File
//void *buff Pointer to the read buffer(NULL:Forward data to the stream)
//word btr Number of bytes to read
//word *br  Pointer to number of bytes read
byte pf_read(void *buff,word btr,word *br)
{
 byte dr;
 CLUST clst;
 dword sect,remain;
 word rcnt;
 byte cs,*rbuff=buff;
 FATFS *fs=FatFs;

 *br=0;
 if (!fs)return FR_NOT_ENABLED;  //Check file system
 if(!(fs->flag & FA_OPENED))return FR_NOT_OPENED; //Check if opened

 remain=fs->fsize-fs->fptr;
 if(btr>remain)btr=(word)remain;   //Truncate btr by remaining bytes

 while (btr){         //Repeat until all data transferred
  if((fs->fptr % 512)==0){    //On the sector boundary?
   cs=(byte)(fs->fptr/512 &(fs->csize-1)); //Sector offset in the cluster
   if(!cs){        //On the cluster boundary?
    clst=(fs->fptr==0)?fs->org_clust:get_fat(fs->curr_clust);   //On the top of the file?
    if(clst<=1)goto fr_abort;
    fs->curr_clust=clst;    //Update current cluster
   }
   sect=clust2sect(fs->curr_clust);  //Get current sector
   if(!sect)goto fr_abort;
   fs->dsect=sect+cs;
  }
  rcnt=(word)(512-(fs->fptr%512));  //Get partial sector data from sector buffer
  if(rcnt>btr)rcnt=btr;
  dr=disk_readp(!buff?0:rbuff,fs->dsect,(word)(fs->fptr%512),rcnt);
  if(dr)goto fr_abort;
  fs->fptr+=rcnt;
  rbuff+=rcnt;   //Update pointers and counters
  btr-=rcnt;
  *br+=rcnt;
 }
 return FR_OK;

fr_abort:
 fs->flag=0;
 return FR_DISK_ERR;
}
//############################################################################//
